"""
All components and wrappers currently
supported by rewx.
"""
import wx
import wx.adv
import wx.lib.scrolledpanel
import wx.richtext
import wx.media
import wx.html
import wx.html2

ActivityIndicator = wx.ActivityIndicator
Button = wx.Button
BitmapButton = wx.BitmapButton
CalendarCtrl = wx.adv.CalendarCtrl
CheckBox = wx.CheckBox
# CollapsiblePane = wx.CollapsiblePane
ComboBox = wx.ComboBox
Dropdown = ComboBox
Frame = wx.Frame
Gauge = wx.Gauge
ListBox = wx.ListBox
ListCtrl = wx.ListCtrl
Panel = wx.Panel
RadioBox = wx.RadioBox
RadioButton = wx.RadioButton
ScrolledPanel = wx.lib.scrolledpanel.ScrolledPanel
Slider = wx.Slider
SpinCtrl = wx.SpinCtrl
SpinCtrlDouble = wx.SpinCtrlDouble
StaticBitmap = wx.StaticBitmap
StaticBox = wx.StaticBox
StaticLine = wx.StaticLine
StaticText = wx.StaticText
TextCtrl = wx.TextCtrl
ToggleButton = wx.ToggleButton
MediaCtrl = wx.media.MediaCtrl
RichTextCtrl = wx.richtext.RichTextCtrl
HtmlWindow = wx.html.HtmlWindow
WebView = wx.html2.WebView


class Grid(wx.Panel):
    """
    Wrapper type for creating a panel
    with a Grid Sizer
    """
    pass

class Block(wx.Panel):
    """
    Wrapper type for creating a panel
    with a BoxSizer
    """

class TextArea(wx.TextCtrl):
    """
    TextCtrl Wrapper with it's style pre-baked
    to be TE_MULTILINE
    """
    pass

class SVG(wx.StaticBitmap):
    """
    Wrapper for converting an SVG to a scaled StaticBitmap
    """
    pass

class SVGButton(wx.BitmapButton):
    """
    Wrapper for creating an SVG backed BitmapButton.
    """
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


import threading
import time
import wx
from concurrent.futures import ThreadPoolExecutor
from random import randint
from typing import Tuple, Callable, Any

from components import StaticText
from rewx import Component, wsx, render, create_element
from rewx.components import Block, Button, Gauge


def fake_download(args: Tuple[str, Callable[[Any], None]]):
    id, update_fn = args
    # pick an arbitrary amount of seconds for the "download" to take
    duration = randint(3, 7)
    # just to make it more visually exciting, we chunk each second
    # into 10 sub-steps
    duration_steps = duration * 10
    step_size = 100 / duration_steps

    for i in range(duration_steps+1):
        time.sleep(0.01)
        # CallAfter is used to safely send an action to the
        # main WX Thread
        wx.CallAfter(update_fn, item={
            'id': id,
            'percent': step_size * i,
            'status': 'Downloading',
            'eta': duration,
        })
    # Gauge's display lags behind its actual value because of
    # how it animates. So, we give a final sleep here after
    # completing our fake work just so that gauge's animation has
    # time to catch back up.
    time.sleep(1)



class FooGauge(Component):
    def __init__(self, props):
        super().__init__(props)
        self.props = props
        self.state = {
            "download_1": 0,
            "download_2": 0,
            # in practice, these downloads would be stored in
            # a proper data structure. They're stored flatly here
            # just for example.
            'status': 'READY'
        }

    def finish_download(self):
        wx.MessageDialog(None, "Download complete!", 'Alert').ShowModal()
        self.set_state({**self.state, 'status': 'READY'})

    def start_download(self, event):
        self.set_state({**self.state, "status": 'DOWNLOADING'})
        thread = threading.Thread(target=self.start_fake_downloads)
        thread.start()

    def update_downloads(self, item):
        self.set_state({
            **self.state,
            item['id']: item['percent']
        })

    def start_fake_downloads(self):
        with ThreadPoolExecutor() as executor:
            args = [
                ('download_1', self.update_downloads),
                ('download_2', self.update_downloads)]
            results = list(executor.map(fake_download, args))
            wx.CallAfter(self.finish_download)


    def render(self):
        return wsx(
            [Block, {},
             [Block, {'orient': wx.HORIZONTAL},
              [StaticText, {'label': 'Download #1'}],
              [Gauge, {"value": self.state["download_1"],
                       "range": 100}]],
             [Block, {'orient': wx.HORIZONTAL},
              [StaticText, {'label': 'Download #2'}],
              [Gauge, {"value": self.state["download_2"],
                       "range": 100}]],
             [Button, {"label": "Download" if self.state['status'] == 'READY' else 'In Progress',
                       'enabled': self.state['status'] == 'READY',
                       "on_click": self.start_download,
                       "flag": wx.CENTER | wx.ALL}],
            ]
        )


if __name__ == "__main__":
    app = wx.App()

    frame = wx.Frame(None, title="Gauge With Update")
    body = render(create_element(FooGauge, {}), frame)

    frame.Show()
    app.MainLoop()


"""
https://medium.com/@sweetpalma/gooact-react-in-160-lines-of-javascript-44e0742ad60f
"""
import functools
import wx
from inspect import isclass

from rewx.dispatch import mount, update
from rewx.widgets import mount as _mount
from rewx.widgets import update as _update

mount.merge_registries(_mount._registry)
update.merge_registries(_update._registry)



def wsx(f):
    def convert(spec: list):
        type, props, *children = spec
        return create_element(type, props, children=list(map(convert, children)))
    # being used as a decorator
    if callable(f):
        @functools.wraps(f)
        def inner(*args, **kwargs):
            result = f(*args, **kwargs)
            return convert(result)
        return inner
    else:
        return convert(f)


def create_element(type, props, children=None):
    element = {
        'type': type,
        'props': props
    }
    if children:
        if not isinstance(children, list):
            raise Exception('Children must be a list!')
        element['props']['children'] = children
    return element



def updatewx(instance, props):
    if isinstance(instance, wx.StaticText):
        instance: wx.StaticText = instance
        if props.get('on_click'):
            instance.Unbind(wx.EVT_LEFT_DOWN)
            instance.Unbind(wx.EVT_LEFT_DCLICK)
            instance.Bind(wx.EVT_LEFT_DOWN, props.get('on_click'))
            instance.Bind(wx.EVT_LEFT_DCLICK, props.get('on_click'))
        else:
            instance.Unbind(wx.EVT_LEFT_DCLICK)
            instance.Unbind(wx.EVT_LEFT_DOWN)
        instance.SetLabel(props.get('value', ''))
    elif isinstance(instance, wx.Panel):
        instance: wx.Panel = instance
        sizer: wx.BoxSizer = instance.GetSizer()
        sizer.SetOrientation(props.get('orient', wx.VERTICAL))
    return instance


def patch(dom: wx.Window, vdom):
    parent = dom.GetParent()
    try:
        # if parent:
        #     parent.Freeze()
        if not isclass(vdom['type']):
            # because stateless functions are just opaque wrappers
            # they have no relevant diffing logic -- there is no
            # associated top-level WX element produced from a SFC, only
            # their inner contents matter. As such, we evaluate it and
            # push the result back into `patch`
            return patch(dom, vdom['type'](vdom['props']))
        if isclass(vdom['type']) and issubclass(vdom['type'], Component):
            return Component.patch_component(dom, vdom)
        elif not isinstance(dom, vdom['type']):
            for child in dom.GetChildren():
                dom.RemoveChild(child)
                child.Destroy()
            dom.Destroy()
            newdom = render(vdom, parent)
        elif isinstance(dom, vdom['type']):
            update(vdom, dom)
            pool = {f'__index_{index}': child for index, child in enumerate(dom.GetChildren())}
            for index, child in enumerate(vdom['props'].get('children', [])):
                key = f'__index_{index}'
                if key in pool:
                    patch(pool[key], child)
                    del pool[key]
                else:
                    # TODO: this IS the addition case, right?
                    # why would I need this removeChild line..?
                    if key in pool:
                        # if we're adding something new to the
                        # tree, it won't be present in the pool
                        parent.RemoveChild(pool[key])
                    # TODO: need to understand this case more
                    # if we're not updating, we're adding
                    # in which case.. why doesn't this fall to the
                    # `dom` instance..?
                    inst = render(child, dom)
                    if dom.GetSizer():
                        dom.GetSizer().Add(
                            inst,
                            child['props'].get('proportion', 0),
                            child['props'].get('flag', 0),
                            child['props'].get('border', 0)
                        )
            # any keys which haven't been removed in the
            # above loop represent wx.Objects which are no longer
            # part of the virtualdom and should thus be removed.
            for key, orphan in pool.items():
                dom.RemoveChild(orphan)
                orphan.Destroy()

            newdom = dom
        else:
            raise Exception("unexpected case!")
        p = parent
        while p:
            p.Layout()
            p = p.GetParent()
        return newdom
    finally:
        # TODO: we sometimes call parent.Thaw() when
        # parent isn't frozen. I think this has something
        # to do with the child removal case. Not sure tho
        # if parent and parent.IsFrozen():
        #     parent.Thaw()
        pass


class Component:
    def __init__(self, props):
        self.props = props
        self.state = None
        # this gets set dynamically once mounted / instantiated
        self.base = None

    @classmethod
    def render_component(cls, vdom, parent=None):
        if cls.__name__ == vdom['type'].__name__:
            instance = vdom['type'](vdom['props'])
            instance.base = render(instance.render(), parent)
            instance.base._instance = instance
            instance.base._key = vdom['props'].get('key', None)
            instance.component_did_mount()

            return instance.base
        else:
            # TODO: what are the cases where this would be hit..?
            return render(vdom['type'](vdom['props']), parent)

    @classmethod
    def patch_component(cls, dom, vdom):
        parent = dom.GetParent()
        # TODO: is any of this right..?
        if hasattr(dom, '_instance') and type(dom._instance).__name__ == vdom['type'].__name__:
            return patch(dom, dom._instance.render())
        if cls.__name__ == vdom['type'].__name__:
            return cls.render_component(vdom, parent)
        else:
            return patch(dom, vdom['type'](vdom['props']))




    def component_did_mount(self):
        pass

    def component_will_unmount(self):
        print('gooodbye!')
        pass

    def render(self):
        return None

    def set_state(self, next_state):
        prev_state = self.state
        self.state = next_state
        p = self.base
        while p.GetParent() != None:
            p = p.GetParent()
        p.Freeze()
        patch(self.base, self.render())
        p.Thaw()



def render(element, parent):
    if isclass(element['type']) and issubclass(element['type'], wx.Object):
        instance = mount(element, parent)
        if element['props'].get('ref'):
            element['props'].get('ref').update_ref(instance)
        for child in element['props'].get('children', []):
            sizer = instance.GetSizer()
            if not sizer:
                render(child, instance)
            else:
                sizer.Add(
                    render(child, instance),
                    child['props'].get('proportion', 0),
                    child['props'].get('flag', 0),
                    child['props'].get('border', 0)
                )
        if isinstance(instance, wx.Frame):
            instance.Layout()
        return instance
    elif type(element['type']) == type:
        return element['type'].render_component(element, parent)
    elif callable(element['type']):
        # stateless functional component
        return render(element['type'](element['props']), parent)
    else:
        # TODO: rest of this message
        raise TypeError(f'''
            An unknown type ("{element['type']}") was supplied as a renderable 
            element. 
        ''')

class Ref:
    def __init__(self):
        self.instance = None

    def update_ref(self, instance):
        self.instance = instance


if __name__ == '__main__':
    statictext = wx.StaticText

    foo_elm = create_element('block', {}, children=[
        create_element('statictext', {'value': 'Hey there, world!'}),
        create_element('statictext', {'value': 'Hey there, again!'}),
        create_element('block', {'orient': wx.HORIZONTAL}, children=[
            create_element('statictext', {'value': 'One'}),
            create_element('statictext', {'value': ' and Two!'}),
        ])
    ])

    foo_elm1 = create_element('block', {}, children=[
        create_element('statictext', {'value': 'One'}),
        create_element('statictext', {'value': 'Two'})
    ])

    foo_elm2 = create_element('block', {'orient': wx.HORIZONTAL}, children=[
        create_element('statictext', {'value': 'Two'}),
        create_element('statictext', {'value': 'One'}),
    ])

    # foo_elm3 = create_element(Foo, {'item1': 'HELLOOOOO'})
    # foo_elm4 = create_element(Bar, {})
    #
    # foo_elm5 = create_element(Bar, {'item1': 'HELLOOOOO'})
    # foo_elm6 = create_element(Foo, {'item1': 'BYeeeee'})

    # basic_app('My Hello App', foo_elm)

    app = wx.App()
    import wx.lib.inspection
    wx.lib.inspection.InspectionTool().Show()
    frame = wx.Frame(None, title='Test re-wx')
    frame.SetSize((570, 520))
    thing = render(create_element(statictext, {'label': 'Two'}), frame)
    # thing = patch(thing, foo_elm6)
    # t = Thread(target=andthen, args=(thing, foo_elm6))
    # t.start()
    box = wx.BoxSizer(wx.VERTICAL)
    box.Add(thing, 1, wx.EXPAND)
    frame.SetSizer(box)
    frame.Show()
    # frame.Fit()

    for child in frame.GetChildren():
        for ccc in child.GetChildren():
            for cc in ccc.GetChildren():
                cc.Layout()
            ccc.Layout()
        child.Layout()
    app.MainLoop()





import wx
from unittest import TestCase

from rewx import create_element, wsx, render
from rewx.components import Gauge, Frame


class TestGaugeElement(TestCase):

    def test_gauge(self):
        expected_state = {
            'value': 123,
            'range': 3000,
            'name': 'Counter',
            'size': (500, 1),
        }
        app = wx.App()
        component = wsx([Frame, {'show': True}, [Gauge, expected_state]])
        frame = render(component, None)
        gauge = frame.Children[0]

        self.assertEqual(gauge.Size, expected_state['size'])
        self.assertEqual(gauge.Range, expected_state['range'])
        self.assertEqual(gauge.Name, expected_state['name'])
        self.assertEqual(gauge.Value, expected_state['value'])





from unittest import TestCase

from rewx import create_element, wsx


class TestCreateElement(TestCase):

    def test_create_element(self):

        element = create_element('foo', {'bar': 'baz'})
        self.assertEqual(element, {
            'type': 'foo',
            'props': {'bar': 'baz'},
        })
        # children key not present in props if not supplied
        self.assertNotIn('children', element['props'])


        # when supplied, `children` get added to props
        element = create_element('foo', {'bar': 'baz'}, children=[
            create_element('some-elm', {'some': 'value'})
        ])
        self.assertEqual(element, {
            'type': 'foo',
            'props': {
                'bar': 'baz',
                'children': [{
                    'type': 'some-elm',
                    'props': {'some': 'value'}
                }]
            }
        })

    def test_wsx(self):
        # calling as function
        result = wsx(['foo', {'bar': 'baz'}])
        self.assertEqual(result, create_element('foo', {'bar': 'baz'}))


        @wsx
        def my_component(props):
            return ['foo', props]

        # asserting the function name gets @wrapped correctly
        self.assertEqual(my_component.__name__, 'my_component')

        self.assertEqual(my_component({'bar': 'baz'}), create_element('foo', {'bar': 'baz'}))


    def test_too(self):
        #TODO: some of these cases can only be tested once things get mounted
        @wsx
        def comp_b(props):
            return ['b', {'bar': props['from_comp_a']}, props.get('children')]

        @wsx
        def comp_a(props):
            return [comp_a, {'from_comp_a': 'hello!'},
                    ['child1', {}],
                    ['child2', {}]]

        print(comp_a({'uhh':'what'}))


import wx


def dispatch(func):
    """
    Constructs a registry for performing single dispatch based on
    the class (not instance) type
    """
    registry = {'default': func}

    def merge_registries(other):
        registry.update(other)

    def register(cls, func=None):
        """
        """
        assert issubclass(cls, wx.Object), "TODO! Explain why it must be one of these!"

        if func is None:
            return lambda f: register(cls, f)
        registry[cls] = func
        return func

    def wrapper(element, parent: wx.Window):
        try:
            return registry[element['type']](element, parent)
        except KeyError as e:
            return registry['default'](element, parent)

    registry[object] = func
    wrapper.register = register
    wrapper._registry = registry
    wrapper.merge_registries = merge_registries
    wrapper.dispatch = dispatch
    return wrapper


@dispatch
def mount(element, parent):
    raise TypeError(f'''
    Encountered an unknown Type ({element['type']}) while trying to mount. 

    cases: 
     1. typo of rewx type: fix yo' shiz 
     2. that wx-widget isn't supported yet. Plz open issue 
     3. they're trying to use their own component: Register mount and update handlers 
    ''')


@dispatch
def update(element, instance):
    raise TypeError(f'''
        Encountered an unknown Type ({element['type']}) while trying to update 
        a mounted component 

        cases: 
         3. It must be that they're trying to use their own component, but didn't 
          register an update handler: register a handler via @update.register(YOUR_TYPE) 
        ''')
